SystemOrganization addCategory: #'AmbientTalkST-Core'!

Process subclass: #ATActor
	instanceVariableNames: 'mailbox terminate'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST-Core'!

!ATActor class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 11:54'!
new
	"Create a new actor and start it."
	| newActor |
	newActor := super new.
	newActor suspendedContext: [newActor start] asContext.
	newActor priority: Processor activePriority.
	newActor resume.
	^ newActor! !

!ATActor methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 17:08'!
enqueue: aMessage
	"Directly put a message into the actors mailbox."
	mailbox nextPut: aMessage! !

!ATActor methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/31/2011 15:09'!
evaluateAsync: aBlock
	"This is an implementation level method.
	 The block is by convention not referring to
	 unsafe shared state."
	| msg |
	msg := ATMessage new.
	msg receiver: aBlock.
	msg setSelector: #value arguments: #().

	mailbox nextPut: msg.
	
	^ msg promise.! !

!ATActor methodsFor: 'initialize-release' stamp: 'StefanMarr 1/2/2012 11:51'!
initialize
	super initialize.
	mailbox := SharedQueue new.
	
	terminate := false.! !

!ATActor methodsFor: 'initialize-release' stamp: 'StefanMarr 1/1/2012 20:52'!
processIncomingMessages
	| msg result |
	msg := mailbox next.
	result := msg send.
	
	msg result: result! !

!ATActor methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 11:50'!
start
	[terminate] whileFalse:
		[ self processIncomingMessages. ]! !

Object subclass: #ATFarReferenceBase
	instanceVariableNames: 'obj actor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST-Core'!

!ATFarReferenceBase commentStamp: 'StefanMarr 1/1/2012 16:45' prior: 0!
ATFarReferenceBase is providing the data and #cannotInterpret: for ATFarReference.
Since #cannotInterpret: is always send to the super class, we needed this base class.
Refer to ATFarReference for relevant class-side methods.

Instance Variables:
	obj	   - the object this far-reference is pointing to
	actor  - the actor the referenced object is contained in!

ATFarReferenceBase subclass: #ATFarReference
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST-Core'!

!ATFarReference commentStamp: 'StefanMarr 1/1/2012 16:51' prior: 0!
ATFarReference implements a far-reference, i.e., a proxy for an object that is referenced from another actor.
The base class ATFarReferenceBase provides the proxy implementation.
The methodDict of ATFarReference is set to nil to be able to use the #cannotInterpet: mechanism, which enables us
to intercept all messages sent to an object, including #doesNotUnderstand.

See the class-side methods to instantiate a far-reference.!

!ATFarReference class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 21:45'!
deStratify
	self methodDict: self emptyMethodDictionary.! !

!ATFarReference class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 13:50'!
for: anObj in: anActor
	| ref |
	"SmallIntegers and bools are immutable and directly used."
	(((((anObj class == ATFarReference)
	  or: [anObj isKindOf: SmallInteger])
	  or: [anObj isKindOf: Boolean])
	  or: [anObj isKindOf: Character])
	  or: [anObj isNil])
		ifTrue: [^anObj].
	
	"Strings are copied as by-value objects kind of isolates"
	(anObj isKindOf: String)
		ifTrue: [^ anObj copy].
		
	ref := ATFarReferenceBase new.
	ref object: anObj.
	ref actor:  anActor.
	
	ATFarReference adoptInstance: ref.
	
	^ ref! !

!ATFarReference class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 14:03'!
initialize
	"self initialize."

	super initialize.
	self stratify.! !

!ATFarReference class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 22:06'!
isStratified
	^ methodDict isNil! !

!ATFarReference class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 13:07'!
stratify
	self methodDict: nil.! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 16:22'!
actor: anActor
	actor := anActor! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 14:18'!
cannotInterpret: aMessage
	| msg |
	[Transcript cr; show: '#ci: rcvr: ', obj asString, ' msg: ', aMessage asString].

	"In case we have a far reference to a local object, we do it directly.
	 THIS SHOULD NEVER HAPPEN!!!!!!"
	actor == Processor activeProcess ifTrue: [
		Transcript show: 'THIS SHOULD NEVER HAPPEN msg:', aMessage asString.
		^ msg sendTo: obj.
	].
	
	aMessage selector = #unwrapATRef   ifTrue: [^ obj   ].
	aMessage selector = #unwrapATActor ifTrue: [^ actor ].
	
	msg := ATMessage from: aMessage.
	msg receiver: obj.
	msg unwrapFarReferencesFor: actor.
	actor enqueue: msg.
	
	^ ATFarReference for: msg result in: actor.! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 13:07'!
doesNotUnderstand: aMessage 
	"Try this for debugging, and for use in non-stratified mode."
	^ self cannotInterpret: aMessage! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 16:22'!
object: anObj
	obj := anObj! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 13:59'!
unwrapATActor
	^ actor! !

!ATFarReferenceBase methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 13:46'!
unwrapATRef
	^ obj! !

Object subclass: #ATIsolate
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST-Core'!

Object subclass: #ATObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST-Core'!

Object subclass: #ATPromise
	instanceVariableNames: 'signal value'
	classVariableNames: 'NoValueDelivered'
	poolDictionaries: ''
	category: 'AmbientTalkST-Core'!

!ATPromise class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/31/2011 14:20'!
example
	"This is not a real test case, but just to see how it is supposed to work"
	| p |
	p := ATPromise new.
	
	[|val| 
	 Transcript log: 'About to read promised value.'.
	 val := p read.
	 Transcript log: 'Val: ', val asString.] fork.

	Processor yield.
	
	Transcript log: 'About to deliver promised value.'.
	p deliver: 42.
	Transcript log: 'Deliver promised value.'.
	! !

!ATPromise class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/4/2011 15:53'!
initialize
	NoValueDelivered := Object new.! !

!ATPromise methodsFor: 'announcing' stamp: 'StefanMarr 11/4/2011 16:07'!
deliver: aValue
	"Deliver the promised value."
	
	
	"Warning: this is racy, but we don't care for the moment.
	 Should be a single produce anyway. Everything else is a bug.
	 And, we do at least best effort error notification..."
	value = NoValueDelivered ifFalse: [
		| err |
		err := Error new.
		err messageText: 'Promise has been delivered more than once.'.
		err signal.
	].

	value := aValue.
	
	signal signal.
	
	^ value.! !

!ATPromise methodsFor: 'initialization' stamp: 'StefanMarr 11/4/2011 16:09'!
initialize
	value := NoValueDelivered.
	signal := Semaphore new.! !

!ATPromise methodsFor: 'accessing' stamp: 'StefanMarr 11/4/2011 16:01'!
read
	[value = NoValueDelivered] whileTrue: [
		"This is ugly, but I do not see how I can prevent the data
		 race for signals otherwise.
		 In case there are multiple possible consumers everyone
		 but the first will starve when they arrive before the
		 promise was delivered.
		 Now, we just have a delay, but it is guaranteed to return."
		signal waitTimeoutMSecs: 100.
	].
	^ value! !

Message subclass: #ATMessage
	instanceVariableNames: 'sender receiver result'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AmbientTalkST-Core'!

!ATMessage class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 17:06'!
from: aMessage
	"Used to convert a normal Message object to an ATMessage object"
	| newATMsg |
	newATMsg := self new.
	newATMsg setSelector: aMessage selector arguments: aMessage arguments.
	newATMsg lookupClass: aMessage lookupClass.
	^ newATMsg.! !

!ATMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/31/2011 15:11'!
initialize
	super initialize.
	result := ATPromise new.! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:20'!
promise
	^ result! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:15'!
receiver
	^ receiver! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:15'!
receiver: anObj
	receiver := anObj! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 1/1/2012 21:48'!
result
	"Result is blocking on the promise.
	 Use #promise to obtain the promise directly."
	^ result read! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:20'!
result: anObj
	result deliver: anObj! !

!ATMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/1/2012 20:38'!
send
	"answer the result of sending this message to its receiver"

	^ receiver perform: selector withArguments: args! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:15'!
sender
	^ sender! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/25/2011 17:15'!
sender: anObj
	sender := anObj! !

!ATMessage methodsFor: 'as yet unclassified' stamp: 'StefanMarr 1/2/2012 14:00'!
unwrapFarReferencesFor: anActor
	args withIndexDo: [:val :i |
		((val class == ATFarReference) and: [val unwrapATActor == anActor])
			ifTrue: [args at: i put: val unwrapATRef].
	].! !
ATFarReference initialize!
ATPromise initialize!
